{
lightBASS
---------
(c)2003,2008 S.V.Bulba
http://bulba.at.kz/
vorobey@mail.khstu.ru

Description:
------------
Dinamycally loads/unloads BASS.DLL (BASSWMA.DLL, BASS_APE.DLL, BASSFLAC.DLL)
Performs all required checks before calling BASS DLLs
Uses minimal set of constants, types and declarations from
original BASS.PAS, BASSWMA.PAS, BASS_APE.PAS and BASSFLAC.PAS

Written for using with BASS.DLL version 2.4
}

unit lightBASS;

interface

uses
 Windows,Messages,SysUtils;

type
 HMUSIC = DWORD;
 HSAMPLE = DWORD;
 HSTREAM = DWORD;
 HSYNC = DWORD;
 FLOAT = Single;
 QWORD = int64;
 HPLUGIN = DWORD;

 EBASSError = class(Exception);

 SYNCPROC = procedure(handle: HSYNC; channel, data: DWORD; user: Pointer); stdcall;

  BASS_CHANNELINFO = record
    freq: DWORD;
    chans: DWORD;
    flags: DWORD;
    ctype: DWORD;
    origres: DWORD;
    plugin: HPLUGIN;
    sample: HSAMPLE;
    filename: PChar;
  end;

 TBassDll = (BASSDLL,BASSWMADLL,BASSAPEDLL,BASSFLACDLL);

{
Next procedures and functions checks some flags and handles
and if all OK calls BASS.DLL functions. During calling
some errors can be ocurred, all of them are correctly translated
into DELPHI's exceptions with error messages
}

procedure RaiseLastBASSError;

procedure LoadBASS(DLL:TBassDll); //If BASS.DLL is not loaded then loads BASS.DLL,
                        //checks version and gets some procs addresses
                        //if DLL <> BASSDLL then also loads BASS plug-in

procedure UnloadBASS;   //Unload BASS.DLL if it was loaded

procedure InitBASS(device: Integer; freq, flags: DWORD; win: HWND);
                        //calls BASS_Init if BASS was not initialized

procedure FreeBASS;     //calls BASS_Free if BASS was initialized

procedure PlayBASS(FileName:PChar;Stream:boolean;DLL:TBassDll;StartTime,TimeLenMs:integer);
                        //Tries start playing file: Stream = True - as stream,
                        //                          Stream = False - as module.
                        //If successed then set sync for end of music by
                        //WM_PLAYNEXTITEM message

procedure PlayFreeBASS;     //if PlayBASS was OK, stops playing and removes sync

function BASS_StreamCreateFile2(DLL:TBassDll; f: Pointer; flags: DWORD): HSTREAM;

procedure SwitchPause;  //during playing pauses/resumes playback

function GetLengthBASS:QWORD;
                        //returns max position for using with
                        //BASS_ChannelGetPosition.

procedure SetSync;    //Set SYNC_END message
procedure RemoveSync; //Remove SYNC_END message

const
 BASS_NOSOUNDDEVICE = 0;
 BASS_FIRSTSOUNDDEVICE = 1;

//Some consts from BASS.PAS
 BASS_TAG_ID3   = 0;
 BASS_TAG_ID3V2 = 1;
 BASS_TAG_OGG   = 2;

 BASS_TAG_RIFF_INFO  = $100;
 BASS_TAG_MUSIC_NAME = $10000;
 BASS_TAG_MUSIC_MESSAGE = $10001;
 BASS_TAG_MUSIC_INST = $10100;
 BASS_TAG_MUSIC_SAMPLE = $10300;

 BASS_STREAM_PRESCAN     = $20000;
 BASS_STREAM_DECODE      = $200000;

 BASS_SAMPLE_LOOP        = 4;

 BASS_MUSIC_STOPBACK     = $80000;
 BASS_MUSIC_PRESCAN      = BASS_STREAM_PRESCAN;
 BASS_MUSIC_CALCLEN      = BASS_MUSIC_PRESCAN;
 BASS_MUSIC_NOSAMPLE     = $100000;

 BASS_POS_BYTE           = 0;

 BASS_DATA_FFT256   = $80000000;
 BASS_DATA_FFT512   = $80000001;
 BASS_DATA_FFT1024  = $80000002;
 BASS_DATA_FFT2048  = $80000003;
 BASS_DATA_FFT4096  = $80000004;
 BASS_DATA_FFT8192  = $80000005;
 BASS_DATA_FFT_NOWINDOW = $20;

var
//Some BASS.DLL function addresses (see description in original BASS.PAS)
 BASS_GetVersion:function: DWORD; stdcall;
 BASS_ErrorGetCode:function: Integer; stdcall;
 BASS_Init:function (device: Integer; freq, flags: DWORD; win: HWND; clsid: PGUID): BOOL; stdcall;
 BASS_Free:function: BOOL; stdcall;
 BASS_StreamCreateFile:function (mem: BOOL; f: Pointer; offset, length: QWORD; flags: DWORD): HSTREAM; stdcall;
 BASS_StreamFree:function (handle: HSTREAM): BOOL; stdcall;
 BASS_ChannelPlay:function (handle: DWORD; restart: BOOL): BOOL; stdcall;
 BASS_ChannelSetSync:function (handle: DWORD; type_: DWORD; param: QWORD; proc: SYNCPROC; user: Pointer): HSYNC; stdcall;
 BASS_ChannelRemoveSync:function (handle: DWORD; sync: HSYNC): BOOL; stdcall;
 BASS_ChannelPause:function (handle: DWORD): BOOL; stdcall;
 BASS_ChannelStop:function (handle: DWORD): BOOL; stdcall;
 BASS_ChannelGetLength:function (handle, mode: DWORD): QWORD; stdcall;
 BASS_ChannelGetPosition:function (handle, mode: DWORD): QWORD; stdcall;
 BASS_ChannelSetPosition:function (handle: DWORD; pos: QWORD; mode: DWORD): BOOL; stdcall;
 BASS_ChannelGetLevel:function (handle: DWORD): DWORD; stdcall;
 BASS_ChannelGetData:function (handle: DWORD; buffer: Pointer; length: DWORD): DWORD; stdcall;
 BASS_ChannelGetAttribute:function (handle, attrib: DWORD; var value: FLOAT): BOOL; stdcall;
 BASS_ChannelBytes2Seconds:function (handle: DWORD; pos: QWORD): Double; stdcall;
 BASS_ChannelSeconds2Bytes:function (handle: DWORD; pos: Double): QWORD; stdcall;
 BASS_ChannelFlags:function (handle, flags, mask: DWORD): DWORD; stdcall;
 BASS_ChannelGetInfo:function (handle: DWORD; var info:BASS_CHANNELINFO):BOOL;stdcall;
 BASS_ChannelGetTags:function (handle: HSTREAM; tags : DWORD): PChar; stdcall;
 BASS_MusicLoad:function (mem: BOOL; f: Pointer; offset: QWORD; length, flags, freq: DWORD): HMUSIC; stdcall;
 BASS_MusicFree:function (handle: HMUSIC): BOOL; stdcall;

 BASS_WMA_StreamCreateFile:function (mem:BOOL; fl:pointer; offset,length:QWORD; flags:DWORD): HSTREAM; stdcall;
 BASS_APE_StreamCreateFile:function (mem:BOOL; f:Pointer; offset,length:QWORD; flags:DWORD): HSTREAM; stdcall;
 BASS_FLAC_StreamCreateFile:function (mem:BOOL; f:Pointer; offset,length:QWORD; flags:DWORD): HSTREAM; stdcall;

 hiBASS:THandle = 0;       //handle to BASS.DLL instance,
                           //0 => BASS.DLL is not loaded
 hiBASSWMA:THandle = 0;    //handle to BASSWMA.DLL instance,
                           //0 => BASSWMA.DLL is not loaded
 hiBASSAPE:THandle = 0;    //handle to BASS_APE.DLL instance,
                           //0 => BASS_APE.DLL is not loaded
 hiBASSFLAC:THandle = 0;   //handle to BASSFLAC.DLL instance,
                           //0 => BASSFLAC.DLL is not loaded

 MusicHandle:integer = 0;  //handle to stream or module,
                           //0 => no music loaded

 MusicIsStream:boolean;    //True => Music is stream, otherwise module

 BASSPaused:boolean;       //pause flag, used by SwitchPause

 BASSInitialized:boolean = False; //True => BASS_Init was called successfully

 BASSDevice:integer;

implementation

uses MainWin;

const
//Some consts from BASS.PAS
 BASS_SYNC_POS           = 0;
 BASS_SYNC_END                     = 2;
 BASS_MP3_SETPOS         = BASS_STREAM_PRESCAN;

var
 hsEND:HSYNC = 0; //sync handler, used for end of music message (WM_PLAYNEXTITEM)
                  //if <> 0 then sync is set
 hsPOS:HSYNC = 0;                 
 SyncTime,SyncRestart:QWORD;

 CallbackWindow:HWND;

procedure SyncEndProc(handle: HSYNC; channel, data: DWORD; user: Pointer); stdcall;
begin
if (SyncTime = -1) or not Do_Loop or not BASS_ChannelSetPosition(MusicHandle,SyncRestart,BASS_POS_BYTE) then
 PostMessage(CallbackWindow,WM_PLAYNEXTITEM,0,0);
end;

procedure RaiseLastBASSError;
const
 BASSErCodes:array[0..45] of string =
 ('All is OK',
  'Memory error',
  'Can''t open the file',
  'Can''t find a free sound driver',
  'The sample buffer was lost',
  'Invalid handle',
  'Unsupported sample format',
  'Invalid position',
  'BASS_Init has not been successfully called',
  'BASS_Start has not been successfully called',
  'Unknown error',
  'Unknown error',
  'Unknown error',
  'Unknown error',
  'Already initialized/paused/whatever',
  'Unknown error',
  'Unknown error',
  'Unknown error',
  'Can''t get a free channel',
  'An illegal type was specified',
  'An illegal parameter was specified',
  'No 3D support',
  'No EAX support',
  'Illegal device number',
  'Not playing',
  'Illegal sample rate',
  'Unknown error',
  'The stream is not a file stream',
  'Unknown error',
  'No hardware voices available',
  'Unknown error',
  'The MOD music has no sequence data',
  'No internet connection could be opened',
  'Couldn''t create the file',
  'Effects are not enabled',
  'Unknown error',
  'Unknown error',
  'Requested data is not available',
  'The channel is a "decoding channel"',
  'A sufficient DirectX version is not installed',
  'Connection timedout',
  'Unsupported file format',
  'Unavailable speaker',
  'Invalid BASS version',
  'Codec is not available/supported',
  'The channel/file has ended');
var
 ErCode:DWORD;
begin
ErCode := BASS_ErrorGetCode;
if ErCode > 45 then ErCode := 26;
raise EBASSError.Create(BASSErCodes[ErCode])
end;

function TryGet(const p:pointer):pointer;
begin
Result := p;
if p = nil then RaiseLastOSError
end;

procedure LoadBASSPlugIn(var hiBASSplugin:THandle;LibName:string;var Proc:pointer;ProcName:string);
begin
if hiBASSplugin <> 0 then exit;
hiBASSplugin := LoadLibrary(PChar(LibName));
if hiBASSplugin = 0 then
 begin
  FreeLibrary(hiBASS);
  hiBASS := 0;
  raise EBASSError.Create(
    LibName + ' 2.4 required for playing extra files.'#13#10 +
    'Download it from http://bulba.at.kz/');
 end;
try
 Proc := TryGet(GetProcAddress(hiBASSplugin,PChar(ProcName)));
except
 FreeLibrary(hiBASSplugin);
 hiBASSplugin := 0;
 FreeLibrary(hiBASS);
 hiBASS := 0;
 raise;
end;
end;

procedure CheckVersion;
begin
if BASS_GetVersion and $FFFF0000 <> $02040000 then
 raise EBASSError.Create('Sorry, BASS version 2.4 required.')
end;

procedure LoadBASS;
begin
if hiBass = 0 then
 begin
  hiBASS := LoadLibrary('BASS.DLL');
  if hiBASS = 0 then
   raise EBASSError.Create(
    'BASS.DLL 2.4 by Ian Luck required for playing extra file types.'#13#10 +
        'Download it from http://bulba.at.kz/');
  try
   BASS_GetVersion := TryGet(GetProcAddress(hiBASS,'BASS_GetVersion'));
   CheckVersion;
   BASS_ErrorGetCode := TryGet(GetProcAddress(hiBASS,'BASS_ErrorGetCode'));
   BASS_Init := TryGet(GetProcAddress(hiBASS,'BASS_Init'));
   BASS_Free := TryGet(GetProcAddress(hiBASS,'BASS_Free'));
   BASS_StreamCreateFile := TryGet(GetProcAddress(hiBASS,'BASS_StreamCreateFile'));
   BASS_StreamFree := TryGet(GetProcAddress(hiBASS,'BASS_StreamFree'));
   BASS_ChannelPlay := TryGet(GetProcAddress(hiBASS,'BASS_ChannelPlay'));
   BASS_ChannelSetSync := TryGet(GetProcAddress(hiBASS,'BASS_ChannelSetSync'));
   BASS_ChannelRemoveSync := TryGet(GetProcAddress(hiBASS,'BASS_ChannelRemoveSync'));
   BASS_ChannelPause := TryGet(GetProcAddress(hiBASS,'BASS_ChannelPause'));
   BASS_ChannelStop := TryGet(GetProcAddress(hiBASS,'BASS_ChannelStop'));
   BASS_ChannelGetLength := TryGet(GetProcAddress(hiBASS,'BASS_ChannelGetLength'));
   BASS_ChannelGetPosition := TryGet(GetProcAddress(hiBASS,'BASS_ChannelGetPosition'));
   BASS_ChannelSetPosition := TryGet(GetProcAddress(hiBASS,'BASS_ChannelSetPosition'));
   BASS_ChannelGetLevel := TryGet(GetProcAddress(hiBASS,'BASS_ChannelGetLevel'));
   BASS_ChannelGetData := TryGet(GetProcAddress(hiBASS,'BASS_ChannelGetData'));
   BASS_ChannelGetAttribute := TryGet(GetProcAddress(hiBASS,'BASS_ChannelGetAttribute'));
   BASS_ChannelBytes2Seconds := TryGet(GetProcAddress(hiBASS,'BASS_ChannelBytes2Seconds'));
   BASS_ChannelSeconds2Bytes := TryGet(GetProcAddress(hiBASS,'BASS_ChannelSeconds2Bytes'));
   BASS_ChannelFlags := TryGet(GetProcAddress(hiBASS,'BASS_ChannelFlags'));
   BASS_ChannelGetInfo := TryGet(GetProcAddress(hiBASS,'BASS_ChannelGetInfo'));
   BASS_ChannelGetTags := TryGet(GetProcAddress(hiBASS,'BASS_ChannelGetTags'));
   BASS_MusicLoad := TryGet(GetProcAddress(hiBASS,'BASS_MusicLoad'));
   BASS_MusicFree := TryGet(GetProcAddress(hiBASS,'BASS_MusicFree'));
  except
   FreeLibrary(hiBass);
   hiBass := 0;
   raise
  end
 end;
case DLL of
BASSWMADLL:
 LoadBASSPlugIn(hiBASSWMA,'BASSWMA.DLL',@BASS_WMA_StreamCreateFile,'BASS_WMA_StreamCreateFile');
BASSAPEDLL:
 LoadBASSPlugIn(hiBASSAPE,'BASS_APE.DLL',@BASS_APE_StreamCreateFile,'BASS_APE_StreamCreateFile');
BASSFLACDLL:
 LoadBASSPlugIn(hiBASSFLAC,'BASSFLAC.DLL',@BASS_FLAC_StreamCreateFile,'BASS_FLAC_StreamCreateFile');
end;
end;

procedure UnloadBASS;
begin
if hiBASSWMA <> 0 then
 begin
  FreeLibrary(hiBASSWMA);
  hiBASSWMA := 0
 end;
if hiBASS = 0 then exit;
FreeLibrary(hiBASS);
hiBASS := 0;
end;

procedure InitBASS;
begin
if BASSInitialized and (BASSDevice = device) then exit;
FreeBASS;
BASSDevice := device;
CallbackWindow := win;
BASSInitialized := BASS_Init(device,freq,flags,win,nil);
if not BASSInitialized then RaiseLastBASSError;
end;

procedure FreeBASS;
begin
if BASSInitialized then
 begin
  BASSInitialized := False;
  BASS_Free;
 end;
end;

procedure PlayBASS;
var
 f:DWORD;
 Restart:boolean;
begin
if MusicHandle <> 0 then exit;
MusicIsStream := Stream;
f := 0; if Do_Loop then f := BASS_SAMPLE_LOOP;
if Stream then
 begin
 //too slow
{  if Mpeg and ((Length(FileName) <= 2) or (FileName[1] <> '\') or
   (FileName[2] <> '\')) then //exclude prescan if network path
   f := f or BASS_STREAM_PRESCAN;}
  MusicHandle := BASS_StreamCreateFile2(DLL,FileName,f)
 end
else
 MusicHandle := BASS_MusicLoad(False,FileName,0,0,
  BASS_MUSIC_STOPBACK or BASS_MUSIC_PRESCAN {or BASS_MUSIC_POSRESETEX} or f,0);
if MusicHandle = 0 then RaiseLastBASSError;
BASSPaused := False;
Restart := True;
SyncTime := -1;
if StartTime >= 0 then
 begin
  SyncRestart := BASS_ChannelSeconds2Bytes(MusicHandle,StartTime / 1000);
  if not BASS_ChannelSetPosition(MusicHandle,SyncRestart,BASS_POS_BYTE) then
   RaiseLastBASSError
  else
   begin
    Restart := False;
    SyncTime := BASS_ChannelSeconds2Bytes(MusicHandle,(StartTime + TimeLenMs) / 1000);
    SetSync;
   end;
 end;
if Restart and not Do_Loop then SetSync;
if not BASS_ChannelPlay(MusicHandle,Restart) then RaiseLastBASSError
end;

procedure PlayFreeBASS;
begin
if MusicHandle = 0 then exit;
try
 RemoveSync;
 BASS_ChannelStop(MusicHandle);
finally
 if MusicIsStream then
  BASS_StreamFree(MusicHandle)
 else
  BASS_MusicFree(MusicHandle);
 MusicHandle := 0;
end
end;

function BASS_StreamCreateFile2(DLL:TBassDll; f: Pointer; flags: DWORD): HSTREAM;
begin
case DLL of
BASSWMADLL:
 Result := BASS_WMA_StreamCreateFile(False,f,0,0,flags);
BASSAPEDLL:
 Result := BASS_APE_StreamCreateFile(False,f,0,0,flags);
BASSFLACDLL:
 Result := BASS_FLAC_StreamCreateFile(False,f,0,0,flags);
else
 Result := BASS_StreamCreateFile(False,f,0,0,flags);
end; 
end;

procedure SwitchPause;
begin
if MusicHandle = 0 then exit;
if not BASSPaused then
 begin
  BASSPaused := BASS_ChannelPause(MusicHandle);
  if not BASSPaused then RaiseLastBASSError;
 end
else
 begin
  BASSPaused := not BASS_ChannelPlay(MusicHandle,False);
  if BASSPaused then RaiseLastBASSError;
 end
end;

procedure SetSync;
begin
hsEND := BASS_ChannelSetSync(MusicHandle,BASS_SYNC_END,0,SyncEndProc,nil);
if SyncTime <> -1 then
 hsPOS := BASS_ChannelSetSync(MusicHandle,BASS_SYNC_POS,SyncTime,SyncEndProc,nil);
end;

procedure RemoveSync;
begin
 if hsEND <> 0 then
  begin
   if not BASS_ChannelRemoveSync(MusicHandle,hsEND) then RaiseLastBASSError;
   hsEND := 0;
  end;
 if hsPOS <> 0 then
  begin
   if not BASS_ChannelRemoveSync(MusicHandle,hsPOS) then RaiseLastBASSError;
   hsPOS := 0;
  end;
end;

function GetLengthBASS;
begin
Result := 0;
if MusicHandle = 0 then exit;
Result := BASS_ChannelGetLength(MusicHandle,BASS_POS_BYTE);
if int64(Result) = - 1 then RaiseLastBASSError;
end;

end.
